@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2019 The Squid Software Foundation and contributors
+ * Copyright (C) 1996-2023 The Squid Software Foundation and contributors
  *
  * Squid software is distributed under GPLv2+ license and includes
  * contributions from numerous individuals and organizations.
@@ -11,6 +11,7 @@
 #include "client_side.h"
 #include "comm/Connection.h"
 #include "comm/forward.h"
+#include "debug/Messages.h"
 #include "ExternalACLEntry.h"
 #include "http/Stream.h"
 #include "HttpReply.h"
@@ -24,20 +25,20 @@
 CBDATA_CLASS_INIT(ACLFilledChecklist);
 
 ACLFilledChecklist::ACLFilledChecklist() :
-    dst_rdns(NULL),
-    request (NULL),
-    reply (NULL),
+    dst_rdns(nullptr),
+    request (nullptr),
+    reply (nullptr),
 #if USE_AUTH
-    auth_user_request (NULL),
+    auth_user_request (nullptr),
 #endif
 #if SQUID_SNMP
-    snmp_community(NULL),
+    snmp_community(nullptr),
 #endif
 #if USE_OPENSSL
-    sslErrors(NULL),
+    sslErrors(nullptr),
 #endif
     requestErrorType(ERR_MAX),
-    conn_(NULL),
+    conn_(nullptr),
     fd_(-1),
     destinationDomainChecked_(false),
     sourceDomainChecked_(false)
@@ -64,7 +65,7 @@ ACLFilledChecklist::~ACLFilledChecklist()
     cbdataReferenceDone(sslErrors);
 #endif
 
-    debugs(28, 4, HERE << "ACLFilledChecklist destroyed " << this);
+    debugs(28, 4, "ACLFilledChecklist destroyed " << this);
 }
 
 static void
@@ -75,7 +76,7 @@ showDebugWarning(const char *msg)
         return;
 
     ++count;
-    debugs(28, DBG_IMPORTANT, "ALE missing " << msg);
+    debugs(28, Important(58), "ERROR: ALE missing " << msg);
 }
 
 void
@@ -116,7 +117,6 @@ ACLFilledChecklist::verifyAle() const
     if (reply && !al->reply) {
         showDebugWarning("HttpReply object");
         al->reply = reply;
-        HTTPMSGLOCK(al->reply);
     }
 
 #if USE_IDENT
@@ -147,12 +147,20 @@ ACLFilledChecklist::conn() const
 }
 
 void
-ACLFilledChecklist::conn(ConnStateData *aConn)
+ACLFilledChecklist::setConn(ConnStateData *aConn)
 {
-    if (conn() == aConn)
-        return;
-    assert (conn() == NULL);
+    if (conn_ == aConn)
+        return; // no new information
+
+    // no conn_ replacement/removal to reduce inconsistent fill concerns
+    assert(!conn_);
+    assert(aConn);
+
+    // To reduce inconsistent fill concerns, we should be the only ones calling
+    // fillConnectionLevelDetails(). Set conn_ first so that the filling method
+    // can detect (some) direct calls from others.
     conn_ = cbdataReference(aConn);
+    aConn->fillConnectionLevelDetails(*this);
 }
 
 int
@@ -211,20 +219,20 @@ ACLFilledChecklist::markSourceDomainChecked()
  *    checkCallback() will delete the list (i.e., self).
  */
 ACLFilledChecklist::ACLFilledChecklist(const acl_access *A, HttpRequest *http_request, const char *ident):
-    dst_rdns(NULL),
-    request(NULL),
-    reply(NULL),
+    dst_rdns(nullptr),
+    request(nullptr),
+    reply(nullptr),
 #if USE_AUTH
-    auth_user_request(NULL),
+    auth_user_request(nullptr),
 #endif
 #if SQUID_SNMP
-    snmp_community(NULL),
+    snmp_community(nullptr),
 #endif
 #if USE_OPENSSL
-    sslErrors(NULL),
+    sslErrors(nullptr),
 #endif
     requestErrorType(ERR_MAX),
-    conn_(NULL),
+    conn_(nullptr),
     fd_(-1),
     destinationDomainChecked_(false),
     sourceDomainChecked_(false)
@@ -235,9 +243,15 @@ ACLFilledChecklist::ACLFilledChecklist(const acl_access *A, HttpRequest *http_re
     rfc931[0] = '\0';
 
     changeAcl(A);
+    setRequest(http_request);
+    setIdent(ident);
+}
 
-    if (http_request != NULL) {
-        request = http_request;
+void ACLFilledChecklist::setRequest(HttpRequest *httpRequest)
+{
+    assert(!request);
+    if (httpRequest) {
+        request = httpRequest;
         HTTPMSGLOCK(request);
 #if FOLLOW_X_FORWARDED_FOR
         if (Config.onoff.acl_uses_indirect_client)
@@ -247,13 +261,20 @@ ACLFilledChecklist::ACLFilledChecklist(const acl_access *A, HttpRequest *http_re
             src_addr = request->client_addr;
         my_addr = request->my_addr;
 
-        if (request->clientConnectionManager.valid())
-            conn(request->clientConnectionManager.get());
+        if (const auto cmgr = request->clientConnectionManager.get())
+            setConn(cmgr);
     }
+}
 
+void
+ACLFilledChecklist::setIdent(const char *ident)
+{
 #if USE_IDENT
+    assert(!rfc931[0]);
     if (ident)
         xstrncpy(rfc931, ident, USER_IDENT_SZ);
+#else
+    (void)ident;
 #endif
 }
 
